#include "raytracing.h"
#include "util.h"

void Insert(struct node **root, struct Object *obj, double relative_cost );
bool Branch_and_Bound( struct node*, struct node  *, struct node**, double *bound );
void AddChild( struct node *insert_here, struct node **new_child );

// Adding each of the objects to the bounding volume hierarchy.
void abvh(struct Scene *scene)
{
	int i;
	struct Object *obj;

	printf("Building the bounding volume hierarchy... \n");

	scene->root = NULL;

	for( i = 0; i < scene->num_objects; i++ )
    {
		obj = &scene->object[i];
		//if( strcmp( obj->name, "point" ) == 0 ) continue;

        // Take into account the estimated cost of intersecting a ray with
        // this object as we insert it into the existing bvh.

		Insert(&scene->root, obj, 1.0);
    }
	printf( "done.\n" );
}

// This function adds an object to an existing bounding volume hierarchy.
// It first constructs a bounding box for the new object, then finds the
// optimal place to insert it into the hierachy using branch-and-bound, and
// finally inserts it at the determined location, updating all the costs
// associated with the nodes in the hierarchy as a side-effect.
void Insert( struct node **root, struct Object *obj, double relative_cost )
{
	struct AABB box = GetBox( obj );
    double bound = 0.0;
    struct node *n = (struct node*)malloc(sizeof(struct node));
	struct node *insert_here;
    n->bbox   = box;
    n->object = obj;
    n->SA     = SurfaceArea( box );
    n->EC     = 1.0;
    n->SEC    = relative_cost;
    n->AIC    = relative_cost * n->SA;
    n->SAIC   = 0.0; // There are no child volumes yet.
	n->sibling = NULL;
	n->parent = NULL;
	n->child = NULL;


    // If this is the first node being added to the hierarchy,
    // it becomes the root.
    if( *root == NULL ) { *root = n; return; }

    // Look for the optimal place to add the new node.
    // The branch-and-bound procedure will figure out where
    // to put it so as to cause the smallest increase in the
    // estimated cost of the tree.
    bound = Infinity;
	insert_here = NULL;
    Branch_and_Bound( *root, n, &insert_here, &bound );
    
	// Now actually insert the node in the optimal place.
    AddChild( insert_here, &n );

}

// This method determines the optimal location to insert a new node
// into the hierarchy; that is, the place that will create the smallest
// increase in the expected cost of intersecting a random ray with the
// new hierarchy.  This is accomplished with branch-and-bound, which finds
// the same solution as a brute-force search (i.e. attempting to insert
// the new object at *each* possible location in the hierarchy), but
// does so efficiently by pruning large portions of the tree.
bool Branch_and_Bound( 
    struct node  *the_root,     // Root of the tree to be modified.
    struct node  *the_node,     // The node to add to the hierarchy.
    struct node  **best_parent, // The best node to add it to.
    double *bound )      // Smallest bound achieved thus far.
{
    double a_delta;  // AIC increase due to Root bbox expansion. 
    double r_delta;  // AIC increase due to new child of Root.   
    double c_bound;  // bound - a_delta, or less.
	double new_area;
    bool improved = false;
	struct node *c;

    // Compute the increase in the bounding box of the root due 
    // to adding the new object.  This is used in computing     
    // the new area cost no matter where it ends up.            
    struct AABB root_box = the_root->bbox;
    bool expanded = aabbExpand( &root_box, the_node->bbox );

    // Compute the common sub-expression of the new area cost  
    // and the increment over this that would occur if we     
    // made the new object a child of the_root.                    
    if( expanded )
    {
        new_area = SurfaceArea( root_box );
        a_delta  = ( new_area - the_root->SA ) * the_root->SEC;
        c_bound  = *bound - a_delta;
        if( c_bound <= 0.0 ) return false; // Early cutoff. 
        r_delta  = new_area * the_node->EC + the_node->AIC;
    }
	else
    {
		a_delta = 0.0;
        c_bound = *bound;
        r_delta = the_root->SA * the_node->EC + the_node->AIC;
    }

    // See if adding the new node directly to the root of this tree
    // achieves a better bound than has thus far been obtained (via
    // previous searches).  If so, update all the parameters to reflect
    // this better choice.
    if( r_delta < c_bound )
    {
		*bound       = a_delta + r_delta;  // This is < *bound. 
        *best_parent = the_root;
        c_bound     = r_delta;
        improved    = true;
    }

    // Compute the smallest increment in total area cost (over 
    // "common") which would result from adding the new node 
    // to one of the children of the_root.  If any of them obtains
    // a better bound than achieved previously, then the "best_parent"
    // and "bound" parameters will be updated as a side effect.
    for( c = the_root->child; c != NULL; c = c->sibling )
    {
        if( Branch_and_Bound( c, the_node, best_parent, &c_bound ) )
        {
            *bound   = c_bound + a_delta;
            improved = true;
        }
    }

    return improved;
}

// This function inserts the given child node into the bounding volume
// hierarchy and adjusts all the associated costs that are stored in
// the tree.
void AddChild( struct node *insert_here, struct node **new_child )
{
    double old_AIC;
    double increment;
    struct AABB new_volume = (*new_child)->bbox;
	bool expanded = false;
	struct node *n, *prev;

	if( insert_here->child == NULL ) 
    {
        // The current node is a leaf, so we must convert it into
        // an internal node.  To do this, copy its current contents into
        // a new node, which will become a child of this one.
        insert_here->child         = (struct node*)malloc(sizeof(struct node));
        insert_here->child->bbox   = insert_here->bbox;
        insert_here->child->object = insert_here->object;
        insert_here->child->parent = insert_here;

		insert_here->child->EC      = 0.0;
		insert_here->child->SA      = 0.0;
		insert_here->child->AIC     = 0.0;
		insert_here->child->SEC     = 0.0;
		insert_here->child->SAIC    = 0.0;
		insert_here->child->sibling = NULL;
		insert_here->child->child   = NULL;
 
        // Fill in all the fields of the current node, which has just changed
        // into an internal node with a single child.
        insert_here->EC   = 1;
        insert_here->SA   = insert_here->child->SA;
        insert_here->SEC  = insert_here->child->EC;  // There is only one child.
        insert_here->SAIC = insert_here->child->AIC; // There is only one child.
        insert_here->AIC  = insert_here->SA * insert_here->SEC + insert_here->SAIC;
    }

    // Splice the new child into the linked list of children.
    (*new_child)->sibling = insert_here->child;
    (*new_child)->parent  = insert_here;
    insert_here->child = (*new_child);

    // Update the summed external cost and the summed AIC as a result
    // of adding the new child node.  These do not depend on surface
    // area, so we needn't consider any expansion of the bounding volume.
    insert_here->SEC  += (*new_child)->EC;
    insert_here->SAIC += (*new_child)->AIC;

    // Now take bounding volume expansion into account due to the new child.
    
    if( aabbExpand( &insert_here->bbox, new_volume ) )
    {
        expanded = true;
        insert_here->SA = SurfaceArea( insert_here->bbox );
    }

    // Compute new area cost & how much it increased due to new child.
    old_AIC = insert_here->AIC;
    insert_here->AIC = insert_here->SA * insert_here->SEC + insert_here->SAIC;
    increment = insert_here->AIC - old_AIC;

    // Propagate the information upward, in two phases.  The first phase
    // deals with volumes that get expanded.  Once a volume is reached that does
    // not expand, there will be no more expansion all the way to the root.
    n = insert_here;
    while( expanded )
    {
        expanded = false;
        prev = n;
        n = n->parent;
        if( n != NULL && aabbExpand(&n->bbox, new_volume ) )
        {
            expanded  = true;
            old_AIC   = n->AIC;
            n->SA     = SurfaceArea( n->bbox );
            n->SAIC  += increment;
            n->AIC    = n->SA * n->SEC + n->SAIC;
            increment = n->AIC - old_AIC;
        }
    }

    // From this point up to the root there will be no more expansion.
    // However, we must still propagate information upward.
    while( n != NULL )
    {
        n->SAIC += increment;
        n->AIC  += increment;
        n = n->parent;
    }
}

struct node *SkipChildren(struct node *curr)
{
    // Advance the "current" pointer (curr) to the next unvisited node
    // at the same level or higher.  Since the parent has already been
    // visited, when we move "up" we must also move to the next "sibling".
    while( curr != NULL )
    {
        if( curr->sibling != NULL ) { curr = curr->sibling; break; }
        curr = curr->parent;
    }
    return curr;
}

bool Hit( struct Ray *ray, struct AABB *A, double max_dist )
{
    struct Vec3 R = ray->direction;
    struct Vec3 Q = ray->origin;
    double r, s, t;
    double min = 0.0;
    double max = max_dist;

    if( R.x > 0 ) // Looking down positive X-axis.
        {
        if( Q.x > A->X.max ) return false;
        r = 1.0 / R.x;
        s = ( A->X.min - Q.x ) * r; if( s > min ) min = s;
        t = ( A->X.max - Q.x ) * r; if( t < max ) max = t;
        }
    else if( R.x < 0.0 ) // Looking down negative X-axis.
        {
        if( Q.x < A->X.min ) return false;
        r = 1.0 / R.x;
        s = ( A->X.max - Q.x ) * r; if( s > min ) min = s;
        t = ( A->X.min - Q.x ) * r; if( t < max ) max = t;
        }
    else if( Q.x < A->X.min || Q.x > A->X.max ) return false;

    if( min > max ) return false; // Degenerate interval.

    if( R.y > 0 ) // Looking down positive Y-axis.
        {
        if( Q.y > A->Y.max ) return false;
        r = 1.0 / R.y;
        s = ( A->Y.min - Q.y ) * r; if( s > min ) min = s;
        t = ( A->Y.max - Q.y ) * r; if( t < max ) max = t;
        }
    else if( R.y < 0 ) // Looking down negative Y-axis.
        {
        if( Q.y < A->Y.min ) return false;
        r = 1.0 / R.y;
        s = ( A->Y.max - Q.y ) * r; if( s > min ) min = s;
        t = ( A->Y.min - Q.y ) * r; if( t < max ) max = t;
        }
    else if( Q.y < A->Y.min || Q.y > A->Y.max ) return false;

    if( min > max ) return false; // Degenerate interval.

    if( R.z > 0 ) // Looking down positive Z-axis.
        {
        if( Q.z > A->Z.max ) return false;
        r = 1.0 / R.z;
        s = ( A->Z.min - Q.z ) * r; if( s > min ) min = s;
        t = ( A->Z.max - Q.z ) * r; if( t < max ) max = t;
        }
    else if( R.z < 0 ) // Looking down negative Z-axis.
        {
        if( Q.z < A->Z.min ) return false;
        r = 1.0 / R.z;
        s = ( A->Z.max - Q.z ) * r; if( s > min ) min = s;
        t = ( A->Z.min - Q.z ) * r; if( t < max ) max = t;
        }
    else if( Q.z < A->Z.min || Q.z > A->Z.max ) return false;

    // There is a hit if and only if the intersection interval [min,max]
    // is not degenerate.
    return min <= max;
}
