#ifndef __RAYTRACING_INCLUDED__  // Include this file only once.
#define __RAYTRACING_INCLUDED__

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "color.h"
#include "vec3.h"
#include "interval.h"
#include "aabb.h"

#define MAX_NUM_OBJECTS 1000000
#define MAX_IMAGE_WIDTH 4000
#define MAX_IMAGE_HEIGHT 4000

#define    Pi          ( 3.14159265358979 )
#define    TwoPi       ( 2.0 * Pi )
#define    FourPi      ( 4.0 * Pi )
#define    DegToRad    ( Pi / 180.0 ) // Convert degrees to radians.
#define    Infinity    ( 1.0E20 )     // Should suffice for "infinity" in most cases.
#define    MachEps     ( 1.0E-5 )     // A reasonable value for the "machine epsilon".
#define    OnePlusEps  ( 1.0 + MachEps )
#define    OneMinusEps ( 1.0 - MachEps )

struct Pixel 
{
    int r;
    int g;
    int b;
};


struct Camera
{
    struct Vec3     eye;         // Position of eye.
    struct Vec3     lookat;      // The point we are looking toward.
    struct Vec3     up;          // A vector not parallel to the direction of gaze.
    double          vpdist;      // Distance to the view plane.
    struct Interval x_win;       // Horizontal extent of view window (typically [-1,1]).
    struct Interval y_win;       // Vertical extent of view window (typically [-1,1]).
    int             x_res;       // Horizontal image resolution in pixels.
    int             y_res;       // Vertical image resolution in pixels.
};

struct Material           // Surface material for shading.
{         
    struct Color  diffuse;       // Diffuse color.
    struct Color  specular;      // Color of highlights.
    struct Color  emission;      // Emitted light.
    struct Color  ambient;       // Ambient light (from all directions).
    struct Color  reflectivity;  // Weights for refleced light, between 0 and 1.
    struct Color  translucency;  // Weights for refracted light, between 0 and 1.
    double        Phong_exp;     // Phong exponent for specular highlights.
    double        ref_index;     // Refractive index.
};

struct Object
{
	char name[80];
	struct Material material;
	
	// point
	struct Vec3 position;

	// block
	struct Vec3 Min; // Minimum coordinates along each axis.
    struct Vec3 Max; // Maximum coordinates along each axis.

	// sphere
	struct Vec3 center;
    double radius;

	// triangle
	struct Vec3 A;
	struct Vec3 B;
	struct Vec3 C;

};

// The node struct forms all of the nodes in the bounding volume hierarchy,
// which can be a tree with arbitrary branching (i.e. internal nodes can
// have any number of children).
struct node 
{
    struct Object *object;
    struct AABB    bbox;    // The bounding box containing all the children.
    double  EC;      // External cost of this object.
    double  SA;      // Surface area -- can be zero for primitve.
    double  AIC;     // Adjusted internal cost.
    double  SEC;     // Sum of external costs of children.
    double  SAIC;    // Sum of adjusted internal costs of children.
    struct node   *sibling; // The next child of the node that contains this one. 
    struct node   *parent;  // The node one level up, or NULL if this is the root.
    struct node   *child;   // An object or volume nested inside this one.
};

struct Scene
{
	int num_objects;
	struct Object object[MAX_NUM_OBJECTS];
	int num_lights;
	struct Object *light[MAX_NUM_OBJECTS];
	struct Color env;
	struct node *root;
};


struct Ray	// A ray in R3.
{            
    struct Vec3 origin;         // The ray originates from this point.
    struct Vec3 direction;      // Unit vector indicating direction of ray.
	unsigned generation; // How deep in the ray tree.  1 == generated from eye.
};

struct HitInfo	// Records all info at ray-object intersection.
{          
    struct Object *object; // The object that was hit (set by Intersect).
    double  distance;     // Distance to hit (used & reset by Intersect).
    struct Vec3    point;        // ray-object intersection point (set by Intersect).
    struct Vec3    normal;       // Surface normal (set by Intersect).
    struct Ray     ray;          // The ray that hit the surface (set by Cast).
};

#endif
