#include "raytracing.h"

extern bool sphere_intersect( struct Ray *ray, struct HitInfo *hitinfo, struct Object *obj );
extern bool triangle_intersect( struct Ray *ray, struct HitInfo *hitinfo, struct Object *obj );
extern bool block_intersect( struct Ray *ray, struct HitInfo *hitinfo, struct Object *obj );
extern struct node *SkipChildren(struct node *curr);
extern bool Hit( struct Ray *ray, struct AABB *A, double max_dist );

struct Color Trace( struct Ray *ray, struct Scene *scene );
bool Cast( struct Ray *ray, struct HitInfo *hitinfo, struct Scene *scene );

struct Color Shader( struct Scene *scene, struct HitInfo *hit )
{
    //static const double epsilon = 1.0E-4;
	struct Ray ray;
	struct HitInfo otherhit;
	struct Color ambient, diffuse, specular, color, r, emission;
	struct Material *mat;
	struct Vec3 O, P, N, E, R, LightPos, L;
	struct Object *light;
	double e, k;
	int i;

	mat = &hit->object->material;
	if( mat->emission.blue > 0 || mat->emission.blue > 0 || mat->emission.red > 0 )
		return mat->emission;

    
	ambient = mat->ambient;
    diffuse = mat->diffuse;
    specular = mat->specular;
    color= color_mul_color( ambient, diffuse );
    O = hit->ray.origin;
    P = hit->point;
    N = hit->normal;
    E = Unit( vec3Sub( O, P ) );
    R = Unit( vec3Sub( vec3Mul( 2.0 * vec3InnerProduct( E, N ), N ), E ) ); //R = Unit( ( 2.0 * ( E * N ) ) * N - E );
    r = mat->reflectivity;
    e = mat->Phong_exp;
    k = mat->ref_index;

    if( vec3InnerProduct( E, N ) < 0.0 ) N = vec3Minus( N );  // Flip the normal if necessary.

	for( i = 0; i < scene->num_lights; i++ )
    {
		light = scene->light[i];
        emission = light->material.emission;
		LightPos = light->position;

		L = Unit( vec3Sub( LightPos, P ) );	// Vector from hit point to the light source     
		ray.direction = L;
		ray.origin = vec3Add( P, vec3Mul( MachEps, L ) );
		otherhit.distance = 1E+20;
		if( Cast( &ray, &otherhit, scene ) && otherhit.distance < dist( P, LightPos ) ) 
			continue;	// Check if hit point can see the light

		if( vec3InnerProduct( N, L ) > 0 )	// N*L<0 means this point is back to the light
		{
			// Accumulate the diffuse color
			color = colorAdd( color, scalar_mul_color( vec3InnerProduct( N, L ), color_mul_color( emission, diffuse ) ) );
			
			// Accumulate the specular color
			if( e > 0 && vec3InnerProduct( R, L ) > 0 )	
				color = colorAdd( color, scalar_mul_color( pow( vec3InnerProduct( R ,L ), e ), color_mul_color( emission, specular ) ) );	
		}

    }

	// Accumulate the reflectivity color
	ray.direction=R;
	ray.origin= vec3Add( P, vec3Mul( MachEps, R ) );	// ray.origin = P + epsilon * R
	ray.generation=hit->ray.generation+1;	
	color = colorAdd(color, color_mul_color(r, Trace(&ray, scene))); //color += r * Trace(&ray, scene);

	return color;
	
}

bool Intersect( struct Ray *ray, struct HitInfo *hitinfo, struct Object *obj )
{
	if     ( !strcmp( obj->name, "sphere"   ) ) return sphere_intersect  ( ray, hitinfo, obj );
	else if( !strcmp( obj->name, "triangle" ) ) return triangle_intersect( ray, hitinfo, obj );
	else if( !strcmp( obj->name, "block"    ) ) return block_intersect   ( ray, hitinfo, obj );
	return 0;
}

struct node *getNext(struct node *curr)
{
	if( curr->child == NULL ) curr = SkipChildren(curr);
    else curr = curr->child;
	return curr;
}


// Cast finds the first point of intersection (if there is one)
// between a ray and a list of geometric objects.  If no intersection
// exists, the function returns false.  Information about the
// closest object hit is returned in "hitinfo".
bool Cast( struct Ray *ray, struct HitInfo *hitinfo, struct Scene *scene )
{
	/*int found_a_hit = 0;
	int i;

    // Each intersector is ONLY allowed to write into the "HitInfo"
    // structure if it has determined that the ray hits the object
    // at a CLOSER distance than currently recorded in HitInfo.distance.
    // When a closer hit is found, the fields of the "HitInfo" structure
    // are updated to hold the material of the object that was just hit.
    // Return "true" if any object was hit.

	for( i = 0; i < scene->num_objects; i++ )
	{
		if( Intersect( ray, hitinfo, &scene->object[i] ) )
		{
			// We've hit an object, but continue looking as there might be a closer hit.
			hitinfo->ray = *ray; // Save the ray in world coordinates.

			found_a_hit = 1; 
		}
	}
	return found_a_hit;*/
	
    // Walk the bounding volume hierarchy intersecting, descending down a
    // branch if and only if the ray intersects the volume.  When a leaf is
    // reached, intersect the ray with the object found there.
    bool found_a_hit = false;
    double closest_hit = Infinity;
    struct node *curr = scene->root;

    while( curr != NULL )
    {
		if( curr->child == NULL )	// curr is Leaf
        {
            //if( iter.Curr()->object->Intersect( ray, hitinfo ) )
			if(Intersect(ray, hitinfo, curr->object))
            {
                found_a_hit = true;
                closest_hit = hitinfo->distance;
				// We've hit an object, but continue looking as there might be a closer hit.
				hitinfo->ray = *ray; // Save the ray in world coordinates.

            }

			// Get next node	
			curr = getNext(curr);
        }
        else if( Hit( ray, &curr->bbox, closest_hit ) ) curr = getNext(curr);
        else curr = SkipChildren(curr);
    }
    return found_a_hit;
}

struct Color Trace( struct Ray *ray, struct Scene *scene )
{
    struct Color color;
    struct HitInfo hitinfo;      // Holds info to pass to shader.
    hitinfo.distance = Infinity; // Follow the full ray.

	if( ray->generation <= 3 )
	{
		if( Cast( ray, &hitinfo, scene ) )
			color = Shader( scene, &hitinfo );
		else
			color = scene->env;
	}
	else
	{
		// The recursion has bottomed out. 
		color.blue = 0.15;
		color.green = 0.15;
		color.red = 0.15;
	}
        return color;
}

void trace( double ** result, struct Ray *ray, struct Scene *scene )
{
  double *dbl3;
  struct Color color;

  color = Trace( ray, scene);

  dbl3 = malloc( 3*sizeof(double));
  dbl3[0] = color.blue;
  dbl3[1] = color.green;
  dbl3[2] = color.red;
  *result = dbl3;
}

